@@ -66,6 +66,7 @@ ossl_x509_new_from_file(VALUE filename)
     if (!(fp = fopen(RSTRING_PTR(filename), "r"))) {
 	ossl_raise(eX509CertError, "%s", strerror(errno));
     }
+    rb_fd_fix_cloexec(fileno(fp));
     x509 = PEM_read_X509(fp, NULL, NULL, NULL);
     /*
      * prepare for DER...
@@ -692,20 +693,15 @@ ossl_x509_add_extension(VALUE self, VALUE extension)
 static VALUE
 ossl_x509_inspect(VALUE self)
 {
-    VALUE subject = rb_inspect(ossl_x509_get_subject(self));
-    VALUE issuer = rb_inspect(ossl_x509_get_issuer(self));
-    VALUE serial = rb_inspect(ossl_x509_get_serial(self));
-    VALUE not_before = rb_inspect(ossl_x509_get_not_before(self));
-    VALUE not_after = rb_inspect(ossl_x509_get_not_after(self));
-    return rb_sprintf("#<%"PRIsVALUE": subject=%"PRIsVALUE", "
-		      "issuer=%"PRIsVALUE", serial=%"PRIsVALUE", "
-		      "not_before=%"PRIsVALUE", not_after=%"PRIsVALUE">",
-		      RB_OBJ_CLASSNAME(self),
-		      RB_OBJ_STRING(subject),
-		      RB_OBJ_STRING(issuer),
-		      RB_OBJ_STRING(serial),
-		      RB_OBJ_STRING(not_before),
-		      RB_OBJ_STRING(not_after));
+    return rb_sprintf("#<%"PRIsVALUE": subject=%+"PRIsVALUE", "
+		      "issuer=%+"PRIsVALUE", serial=%+"PRIsVALUE", "
+		      "not_before=%+"PRIsVALUE", not_after=%+"PRIsVALUE">",
+		      rb_obj_class(self),
+		      ossl_x509_get_subject(self),
+		      ossl_x509_get_issuer(self),
+		      ossl_x509_get_serial(self),
+		      ossl_x509_get_not_before(self),
+		      ossl_x509_get_not_after(self));
 }
 
 /*
@@ -814,9 +810,6 @@ Init_ossl_x509cert()
      *   cert.sign(root_key, OpenSSL::Digest::SHA256.new)
      *
      */
-
-    eX509CertError = rb_define_class_under(mX509, "CertificateError", eOSSLError);
-
     cX509Cert = rb_define_class_under(mX509, "Certificate", rb_cObject);
 
     rb_define_alloc_func(cX509Cert, ossl_x509_alloc);
